////////////////////////////////////////////////////////////////////// 
// merlin/KAC.cpp 
// (c) 2000-2001 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Friday December 14, 2001
// 
 
#include "KAC.h"
#include "Pedigree.h"
#include "MathGenMin.h"
#include "MathStats.h"

#include <math.h>

KAC::KAC(StringArray & alias) : labels(alias)
   {
   affectionCount = familyCount = positionCount = 0;
   min = NULL;
   max = NULL;
   scores = NULL;
   }

KAC::~KAC()
   {
   if (min != NULL) delete [] min;
   if (max != NULL) delete [] max;
   if (scores != NULL) delete [] scores;
   }

void KAC::LabelPhenotypes(Pedigree & ped,
     bool all, bool pairs, bool qtl, bool deviates)
   {
   pheno.Clear();

   if (all)
      for (int a = 0; a < ped.affectionCount; a++)
         pheno.Add(ped.affectionNames[a] + " [ALL]");

   if (pairs)
      for (int a = 0; a < ped.affectionCount; a++)
         pheno.Add(ped.affectionNames[a] + " [Pairs]");

   if (qtl)
      for (int t = 0; t < ped.traitCount; t++)
         pheno.Add(ped.traitNames[t] + " [QTL]");

   if (deviates)
      for (int t = 0; t < ped.traitCount; t++)
         pheno.Add(ped.traitNames[t] + " [QTL, Mean = 0.0]");
   }

void KAC::Prepare(int affections, int families)
   {
   if (min != NULL) delete [] min;
   if (max != NULL) delete [] max;
   if (scores != NULL) delete [] scores;

   affectionCount = affections;
   familyCount = families;
   positionCount = labels.Length();

   if (affections == 0) return;

   min = new Vector[affections];
   max = new Vector[affections];
   scores = new Matrix[affections];

   for (int i = 0; i < affections; i++)
      {
      min[i].Dimension(families);
      min[i].Zero();
      max[i].Dimension(families);
      max[i].Zero();
      scores[i].Dimension(families, positionCount);
      scores[i].Zero();
      }
  }

void KAC::SkipFamily(int serial)
   {
   for (int i = 0; i < affectionCount; i++)
      {
      min[i][serial] = 0.0;
      max[i][serial] = 0.0;
      // Setting the min and max to zero and equal is all we need!
      // scores[i][serial].Zero();
      }
   }

void KAC::Set(int affection, int family, double lo, double hi, Vector & zScores)
   {
   min[affection][family] = lo;
   max[affection][family] = hi;
   scores[affection][family] = zScores;
   }

double KAC::Evaluate(Vector & parameter)
   {
   if (parameter[0] < -14 || parameter[0] > 3) return 1e100;

   double delta = MakeDelta(parameter[0]);

   double sum = 0.0;

   for (int i = 0; i < familyCount; i++)
      sum += 2.0 * log(1.0 + delta * scores[currentAffection][i][currentPosition]);

   return -sum;
   }

double KAC::MakeDelta(double value)
   {
   return exp(-exp(value)) * range + minimum;
   }

void KAC::Analyse()
   {
   PowellMinimizer powell;
   powell.func = this;
   powell.Reset(1);

   for (currentAffection=0; currentAffection<affectionCount; currentAffection++)
      {
      // Calculate the range of Z-scores and d-hat
      double lo = 0.0, hi = 0.0;
      int    informativeFamilies = 0;

      for (int i = 0; i < familyCount; i++)
         if (min[currentAffection][i] != max[currentAffection][i])
            {
            if (min[currentAffection][i] < lo) lo = min[currentAffection][i];
            if (max[currentAffection][i] > hi) hi = max[currentAffection][i];
            informativeFamilies++;
            }

      if (lo == hi)
         {
         printf("\nNo informative affected pairs for %s\n\n",
            (const char *) pheno[currentAffection]);
         continue;
         }

      printf("\nPhenotype: %s (%d famil%s)\n"
             "============================================================\n"
             "                 Pos   Zmean  pvalue    delta    LOD  pvalue\n",
            (const char *) pheno[currentAffection], informativeFamilies,
            informativeFamilies == 1 ? "y" : "ies");

      double minDelta = -1.0 / hi;
      double maxDelta = -1.0 / lo;
      double denominator = 1.0 / sqrt(informativeFamilies);

      minimum = minDelta + TINY;
      range   = maxDelta - TINY - minimum;

      // Calculate the minimum & maximum possible scores
      double minZ = 0.0, minChi = 0.0;
      for (int i = 0; i < familyCount; i++)
         {
         minZ   += min[currentAffection][i];
         minChi += 2.0 * log(1 + minimum * min[currentAffection][i]);
         }
      minZ *= denominator;
      OutputScores( "min", minZ, minDelta, minChi);

      double maxZ = 0.0, maxChi = 0.0;
      double maximum = minimum + range;
      for (int i = 0; i < familyCount; i++)
         {
         maxZ   += max[currentAffection][i];
         maxChi += 2.0 * log(1 + maximum * max[currentAffection][i]);
         }
      maxZ *= denominator;
      OutputScores( "max", maxZ, maxDelta, maxChi);

      // Calculate scores for each analysis position
      for (currentPosition=0; currentPosition<positionCount; currentPosition++)
         {
         double sum = 0.0;

         for (int i = 0; i < familyCount; i++)
            sum += scores[currentAffection][i][currentPosition];

         powell.point[0] = 0;
         double zmean    = sum * denominator;
         double chisq    = -powell.Minimize();
         double delta    = MakeDelta(powell.point[0]);

         OutputScores( labels[currentPosition], zmean, delta, chisq);
         }
      }
   }

void KAC::OutputScores(const char * pos, double zscore, double delta, double chisq)
   {
   double pvalue = ndist(zscore);
   double LOD    = chisq * 0.2171472409516;
   double pdelta = chisq > 0 ? chidist(chisq, 1.0) * 0.5 : 0.5;

   if (delta < 0)
      {
      pdelta = 1.0 - pdelta;
      LOD = -LOD;
      }

   int digits  = pvalue < 1.5e-4 ? 5 : int(log(pvalue*0.06666666)*-0.4343);
   int digits2 = pdelta < 1.5e-4 ? 5 : int(log(pdelta*0.06666666)*-0.4343);

   printf("%20s %7.2f %7.*f %8.3f %6.2f %7.*f\n",
          pos, zscore, digits, pvalue, delta, LOD, digits2, pdelta);
   }
 
